/* Special definitions for GNU m4, processed by autoheader.
   Copyright (C) 1993, 1994 Free Software Foundation, Inc.
   Francois Pinard <pinard@iro.umontreal.ca>, 1993.
*/

/* Define to 1 if the changeword(REGEXP) functionnality is wanted.  */
#undef ENABLE_CHANGEWORD

/* Define to 1 if you have ecvt(3), fcvt(3) and gcvt(3), define to 2 if
   these are declared in <stdlib.h>.  */
#undef HAVE_EFGCVT

/* Define to 1 if #include <signal.h> declares struct sigcontext */
#undef HAVE_SIGCONTEXT

/* Define to the name of the distribution.  */
#undef PRODUCT

/* Define to 1 if ANSI function prototypes are usable.  */
#undef PROTOTYPES

/* Define to int if rlim_t is not defined in sys/resource.h */
#undef rlim_t

/* Define to struct sigaltstack if stack_t is not defined in sys/signal.h */
#undef stack_t

/* Define to 1 if using stack overflow detection.  */
#undef USE_STACKOVF

/* Define to the version of the distribution.  */
#undef VERSION

/* Define to 1 for better use of the debugging malloc library.  See 
   site ftp.antaire.com in antaire/src, file dmalloc/dmalloc.tar.gz.  */
#undef WITH_DMALLOC
